import requests

from requests.exceptions import RequestException
from toolz import valfilter, merge

from .utils import escape


_default = dict(host='http://localhost:8123', database='default',
                user=None, password=None)


class ClickhouseException(Exception):
    pass


def execute(query, connection={}, data=None, external={}, stream=False):
    connection = merge(_default, connection)
    params = {'query': query.format(db=escape(connection['database'])),
              'user': connection['user'],
              'password': connection['password']}
    params = valfilter(lambda x: x, params)

    files = {}
    for name, (structure, format, serialized) in external.items():
        params['{}_format'.format(name)] = format
        params['{}_structure'.format(name)] = structure
        files[name] = serialized

    if data is not None:
        data = data.encode('utf-8')
    response = requests.post(connection['host'], params=params, data=data,
                             stream=stream, files=files)

    try:
        response.raise_for_status()
    except RequestException as e:
        if response.content:
            raise ClickhouseException(response.content)
        else:
            raise e

    if stream:
        return response.raw
    else:
        return response.content



import csv

import pandas as pd
from toolz import keymap, valmap

from .http import execute
from .utils import PD2CH, CH2PD, escape


def prepare(df, index=True):
    if index:
        df = df.reset_index()

    for col in df.select_dtypes([bool]):
        df[col] = df[col].astype('uint8')

    dtypes = valmap(PD2CH.get, dict(df.dtypes))
    if None in dtypes.values():
        raise ValueError('Unknown type mapping in dtypes: {}'.format(dtypes))

    return dtypes, df


def serialize(df):
    return df.to_csv(header=False, index=False, quoting=csv.QUOTE_NONNUMERIC,
                     encoding='utf-8')


def read_clickhouse(query, index=True, tables={}, connection={}, **kwargs):
    """Reads clickhouse query to pandas dataframe

    Parameters
    ----------

    query: str
        Clickhouse sql query, {db} will automatically replaced
        with `database` argument
    host: str
        clickhouse host to connect
    tables: dict of pandas DataFrames
        external table definitions for query processing
    database: str, default 'default'
        clickhouse database
    user: str, default None
        clickhouse user
    password: str, default None
        clickhouse password
    index: bool, default True
        whether to serialize `tables` with index or not

    Additional keyword arguments passed to `pandas.read_table`
    """
    query = query.strip().strip(';')
    query = '{} FORMAT TSVWithNamesAndTypes'.format(query)

    external = {}
    for name, df in tables.items():
        dtypes, df = prepare(df, index=index)
        dtypes = keymap(escape, dtypes)
        data = serialize(df)
        structure = ', '.join(map(' '.join, dtypes.items()))
        external[name] = (structure, data)

    lines = execute(query, external=external, stream=True,
                    connection=connection)

    names = lines.readline().decode('utf-8').strip().split('\t')
    types = lines.readline().decode('utf-8').strip().split('\t')
    dtypes, dtimes = {}, []
    for name, chtype in zip(names, types):
        dtype = CH2PD[chtype]
        if dtype.startswith('datetime'):
            dtimes.append(name)
        else:
            dtypes[name] = dtype

    return pd.read_table(lines, header=None, names=names,
                         dtype=dtypes, parse_dates=dtimes,
                         **kwargs)


def to_clickhouse(df, table, index=True, chunksize=1000, connection={}):
    insert = 'INSERT INTO {db}.{table} ({columns}) FORMAT CSV'
    dtypes, df = prepare(df, index=index)

    columns = ', '.join(map(escape, dtypes.keys()))
    query = insert.format(db='{db}', columns=columns, table=escape(table))

    nrows = df.shape[0]
    nchunks = int(nrows / chunksize) + 1
    for i in range(nchunks):
        start_i = i * chunksize
        end_i = min((i + 1) * chunksize, nrows)
        if start_i >= end_i:
            break

        chunk = df.ix[start_i:end_i]
        execute(query, data=serialize(chunk), connection=connection)

    return nrows



import numpy as np
from toolz import itemmap, keymap


MAPPING = {'object': 'String',
           'uint64': 'UInt64',
           'uint32': 'UInt32',
           'uint16': 'UInt16',
           'float64': 'Float64',
           'float32': 'Float32',
           'uint8': 'UInt8',
           'int64': 'Int64',
           'int32': 'Int32',
           'int16': 'Int16',
           'int8': 'Int8',
           'datetime64[D]': 'Date',
           'datetime64[ns]': 'DateTime'}

PD2CH = keymap(np.dtype, MAPPING)
CH2PD = itemmap(reversed, MAPPING)


SPECIAL_CHARS = {
    '\b': '\\b',
    '\f': '\\f',
    '\r': '\\r',
    '\n': '\\n',
    '\t': '\\t',
    '\0': '\\0',
    '\\': '\\\\',
    "'": "\\'"
}


def escape(value, quote='`'):
    if not isinstance(value, (str, bytes)):
        return value
    value = ''.join(SPECIAL_CHARS.get(c, c) for c in value)
    if quote == '`':
        return '`{}`'.format(value)
    elif quote == '\'':
        return '\'{}\''.format(value)
    else:
        return value



#!/usr/bin/env python

from os.path import exists
from setuptools import setup


setup(name='pandahouse',
      version='0.2',
      description='Pandas interface for Clickhouse HTTP API',
      url='http://github.com/kszucs/pandahouse',
      maintainer='Krisztian Szucs',
      maintainer_email='szucs.krisztian@gmail.com',
      license='BSD',
      keywords='',
      packages=['pandahouse'],
      tests_require=['pytest'],
      setup_requires=['pytest-runner'],
      install_requires=['pandas', 'requests', 'toolz'],
      long_description=(open('README.rst').read() if exists('README.rst')
                        else ''),
      zip_safe=False)
